#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <c10/cuda/CUDAGuard.h>

// Define warp and tile sizes
#define WARP_SIZE 32
#define TILE_SIZE 4  // Each warp computes TILE_SIZE outputs simultaneously

// Hybrid kernel: combines benefits of multi-output tiling (minimizing kernel launches) from Kernel 1
// and the simplicity of single-warp dot product from Kernel 2. It uses vectorized loads via float4, __ldg() 
// for read-only caching, and warp-level shuffle reductions to eliminate explicit synchronization.

__global__ void warp_tile_hybrid_kernel(const float* __restrict__ x,
                                          const float* __restrict__ weight,
                                          const float* __restrict__ bias,
                                          float* __restrict__ out,
                                          int in_features,
                                          int out_features) {
    // Determine warp and lane indices
    int warp_id = threadIdx.x / WARP_SIZE;
    int lane_id = threadIdx.x % WARP_SIZE;

    // Each block contains several warps
    int warps_per_block = blockDim.x / WARP_SIZE;

    // Batch index is driven by blockIdx.x
    int batch_idx = blockIdx.x;
    
    // blockIdx.y covers groups of output features, each group has (warps_per_block * TILE_SIZE) outputs
    int group_idx = blockIdx.y;
    int base_out_group = group_idx * (warps_per_block * TILE_SIZE);

    // Each warp computes a tile of TILE_SIZE consecutive outputs
    int out_base = base_out_group + warp_id * TILE_SIZE;
    if (out_base >= out_features) return;

    // Initialize accumulators for each output computed by the warp tile
    float sums[TILE_SIZE];
#pragma unroll
    for (int t = 0; t < TILE_SIZE; t++) {
        sums[t] = 0.0f;
    }

    // Pointer to the current input row
    const float* x_row = x + batch_idx * in_features;

    // For each output in the tile, compute the dot product
#pragma unroll
    for (int t = 0; t < TILE_SIZE; t++) {
        int current_out = out_base + t;
        if (current_out < out_features) {
            // Get the corresponding weight row for current output feature
            const float* w_row = weight + current_out * in_features;

            // Use vectorized loads: float4 accesses (requires in_features to be a multiple of 4 for full efficiency)
            int vec_count = in_features / 4;
            const float4* x_vec = reinterpret_cast<const float4*>(x_row);
            const float4* w_vec = reinterpret_cast<const float4*>(w_row);

            // Each thread in the warp handles a strided portion of the vectorized dot-product
            for (int i = lane_id; i < vec_count; i += WARP_SIZE) {
                float4 x_val = __ldg(x_vec + i);
                float4 w_val = __ldg(w_vec + i);
                sums[t] += x_val.x * w_val.x + x_val.y * w_val.y + x_val.z * w_val.z + x_val.w * w_val.w;
            }

            // Process remaining elements, if any, with scalar loads
            int start = vec_count * 4;
            for (int i = start + lane_id; i < in_features; i += WARP_SIZE) {
                sums[t] += __ldg(x_row + i) * __ldg(w_row + i);
            }
        }
    }

    // Perform warp-level reduction with shuffle instructions for each output in the tile
#pragma unroll
    for (int t = 0; t < TILE_SIZE; t++) {
        for (int offset = WARP_SIZE / 2; offset > 0; offset /= 2) {
            sums[t] += __shfl_down_sync(0xffffffff, sums[t], offset);
        }
    }

    // Lane 0 of each warp writes the final result, adding bias and applying ReLU activation
    if (lane_id == 0) {
        for (int t = 0; t < TILE_SIZE; t++) {
            int current_out = out_base + t;
            if (current_out < out_features) {
                float result = sums[t] + __ldg(bias + current_out);
                out[batch_idx * out_features + current_out] = (result > 0.0f) ? result : 0.0f;
            }
        }
    }
}

// Host function: prepares and launches the hybrid kernel
// Grid configuration: blockIdx.x spans the batch dimension and blockIdx.y spans groups of output features
// Each block is organized with a fixed number of threads such that each warp handles TILE_SIZE outputs

torch::Tensor warp_tile_hybrid_forward(torch::Tensor x,
                                         torch::Tensor weight,
                                         torch::Tensor bias) {
    TORCH_CHECK(x.is_cuda(), "x must be a CUDA tensor");
    TORCH_CHECK(weight.is_cuda(), "weight must be a CUDA tensor");
    TORCH_CHECK(bias.is_cuda(), "bias must be a CUDA tensor");

    int batch_size = x.size(0);
    int in_features = x.size(1);
    int out_features = weight.size(0);

    auto out = torch::empty({batch_size, out_features}, x.options());

    // Tunable parameter: Number of warps per block
    const int warps_per_block = 8;
    const int threads_per_block = warps_per_block * WARP_SIZE;

    // Calculate number of groups along the output feature dimension
    int groups_y = (out_features + (warps_per_block * TILE_SIZE) - 1) / (warps_per_block * TILE_SIZE);

    dim3 grid(batch_size, groups_y);
    dim3 block(threads_per_block);

    cudaStream_t stream = c10::cuda::getCurrentCUDAStream();
    warp_tile_hybrid_kernel<<<grid, block, 0, stream>>>(
        x.data_ptr<float>(),
        weight.data_ptr<float>(),
        bias.data_ptr<float>(),
        out.data_ptr<float>(),
        in_features,
        out_features
    );

    return out;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &warp_tile_hybrid_forward, "Hybrid GEMM with bias and ReLU using warp-level tiling and optimized __ldg() loads (CUDA)");
}
